import os

config_path = os.environ.get(
    "SYFTBOX_CLIENT_CONFIG_PATH", os.path.expanduser("~/.syftbox/client_config.json")
)


def get_config_path() -> str:
    config_path = os.environ.get("SYFTBOX_CLIENT_CONFIG_PATH", None)
    if config_path is None:
        config_path = (
            str(
                input(
                    "Path to your syftbox config.json file (eg: /home/user/SyftBox/config/): "
                )
            )
            + "/"
        )
        os.environ["SYFTBOX_CLIENT_CONFIG_PATH"] = (
            config_path + "/"
        )  # Add a / just in case it's not passed.
    return config_path



from pathlib import Path
from typing import Mapping, Union

from loguru import logger

from syftbox.lib.lib import SyftPermission

DirTree = Mapping[str, "Union[str, DirTree]"]


def create_dir_tree(base_path: Path, tree: DirTree) -> None:
    logger.debug(f"creating tree at {base_path}")
    for name, content in tree.items():
        local_path = base_path / name

        if isinstance(content, str):
            local_path.write_text(content)
        elif isinstance(content, SyftPermission):
            content.save(path=str(local_path))
        elif isinstance(content, dict):
            local_path.mkdir(parents=True, exist_ok=True)
            create_dir_tree(local_path, content)



import hashlib
import os
from enum import Enum
from pathlib import Path

from loguru import logger
from pydantic import BaseModel
from typing_extensions import Optional, Self, Union


class SyftBaseModel(BaseModel):
    def to_dict(self) -> dict:
        # used until we remote Jsonable from the code base
        return self.model_dump(mode="json")

    def save(self, path: Union[str, Path]) -> dict:
        if isinstance(path, Path):
            path = str(path)
        with open(path, "w") as f:
            f.write(self.model_dump_json())
        return self.model_dump(mode="json")

    @classmethod
    def load(cls, filepath: str) -> Self:
        with open(filepath) as f:
            data = f.read()
            return cls.model_validate_json(data)


class FileChangeKind(Enum):
    CREATE: str = "create"
    # READ: str "read"
    WRITE: str = "write"
    # append?
    DELETE: str = "delete"


class FileChange(SyftBaseModel):
    kind: FileChangeKind
    parent_path: str
    sub_path: str
    file_hash: str
    last_modified: float
    sync_folder: Optional[str] = None

    @property
    def kind_write(self) -> bool:
        return self.kind in [FileChangeKind.WRITE, FileChangeKind.CREATE]

    @property
    def kind_delete(self) -> bool:
        return self.kind == FileChangeKind.DELETE

    @property
    def full_path(self) -> str:
        return self.sync_folder + "/" + self.parent_path + "/" + self.sub_path

    @property
    def internal_path(self) -> str:
        return self.parent_path + "/" + self.sub_path

    def hash_equal_or_none(self) -> bool:
        if not os.path.exists(self.full_path):
            return True

        local_file_hash = get_file_hash(self.full_path)
        return self.file_hash == local_file_hash

    def newer(self) -> bool:
        if not os.path.exists(self.full_path):
            return True

        local_last_modified = get_file_last_modified(self.full_path)
        if self.last_modified >= local_last_modified:
            return True

        return False

    def is_directory(self) -> bool:
        return os.path.isdir(self.full_path)

    def read(self) -> Optional[bytes]:
        # if is_symlink(self.full_path):
        #     # write a text file with a syftlink
        #     data = convert_to_symlink(self.full_path).encode("utf-8")
        #     return data
        # else:
        if self.is_directory():
            return None
        with open(self.full_path, "rb") as f:
            return f.read()

    def write(self, data: bytes) -> bool:
        # if its a non private syftlink turn it into a symlink
        # if data.startswith(b"syft://") and not self.full_path.endswith(".private"):
        #     syft_link = SyftLink.from_url(data.decode("utf-8"))
        #     abs_path = os.path.join(
        #         os.path.abspath(self.sync_folder), syft_link.sync_path
        #     )
        #     if not os.path.exists(abs_path):
        #         raise Exception(
        #             f"Cant make symlink because source doesnt exist {abs_path}"
        #         )
        #     dir_path = os.path.dirname(self.full_path)
        #     os.makedirs(dir_path, exist_ok=True)
        #     if os.path.exists(self.full_path) and is_symlink(self.full_path):
        #         os.unlink(self.full_path)
        #     os.symlink(abs_path, self.full_path)
        #     os.utime(
        #         self.full_path,
        #         (self.last_modified, self.last_modified),
        #         follow_symlinks=False,
        #     )

        #     return True
        # else:
        return self.write_to(data, self.full_path)

    def delete(self) -> bool:
        try:
            os.unlink(self.full_path)
            return True
        except Exception as e:
            if "No such file" in str(e):
                return True
            logger.info(f"Failed to delete file at {self.full_path}. {e}")
        return False

    def write_to(self, data: bytes, path: str) -> bool:
        base_dir = os.path.dirname(path)
        os.makedirs(base_dir, exist_ok=True)
        try:
            with open(path, "wb") as f:
                f.write(data)
            os.utime(
                path,
                (self.last_modified, self.last_modified),
                follow_symlinks=False,
            )
            return True
        except Exception as e:
            logger.error(f"failed to write to {path}.")
            logger.exception(e)
            return False


class WriteRequest(BaseModel):
    email: str
    change: FileChange
    is_directory: bool = False
    data: Optional[str] = None


class WriteResponse(BaseModel):
    accepted: bool
    change: FileChange
    status: str


class ListDatasitesResponse(BaseModel):
    datasites: list[str]
    status: str


class ReadResponse(BaseModel):
    change: FileChange
    status: str
    is_directory: bool = False
    data: Optional[str] = None


class ReadRequest(BaseModel):
    email: str
    change: FileChange


class FileInfo(SyftBaseModel):
    file_hash: str
    last_modified: float
    num_bytes: int


class DirState(SyftBaseModel):
    tree: dict[str, FileInfo]
    timestamp: float
    sync_folder: str
    sub_path: str


class DirStateRequest(SyftBaseModel):
    email: str
    sub_path: str


class DirStateResponse(SyftBaseModel):
    sub_path: str
    dir_state: DirState
    status: str


def get_file_last_modified(file_path: str) -> float:
    return os.path.getmtime(file_path)


def get_file_size(file_path: str) -> int:
    return os.path.getsize(file_path)


def get_file_hash(file_path: str) -> str:
    # if is_symlink(file_path):
    #     # return the hash of the syftlink instead
    #     sym_link_string = convert_to_symlink(file_path)
    #     return hashlib.md5(sym_link_string.encode("utf-8")).hexdigest()
    # else:
    # TODO: we will run out of memory for very large files
    with open(file_path, "rb") as file:
        return hashlib.md5(file.read()).hexdigest()



from pathlib import Path

from typing_extensions import Union

DEFAULT_WORKSPACE_DIR = Path("~/.syftbox")


class SyftWorkspace:
    """
    A Syft workspace is a directory structure for everything stored by the client.
    Each workspace is expected to be unique for a client.

    syft_root_dir/
    ├── config/                      <-- syft client configuration
    │   └── config.json
    ├── plugin/                      <-- workspace for plugins to store data
    │   └── sync/
    │       └── changelog.txt
    └── sync/                        <-- everything under this gets sync'd
        ├── apps/
        │   └── fedflix
        └── datasites/
            ├── alice@acme.org
            └── bob@acme.org
    """

    def __init__(self, root_dir: Union[Path, str] = DEFAULT_WORKSPACE_DIR):
        self.root_dir = Path(root_dir).expanduser()

        # config dir
        self.config_dir = self.root_dir / "config"

        # plugins dir
        self.plugins_dir = self.root_dir / "plugins"

        # sync dirs
        self.sync_dir = self.root_dir / "sync"
        self.datasites_dir = self.sync_dir / "datasites"
        self.apps_dir = self.sync_dir / "apps"

    def mkdirs(self):
        self.config_dir.mkdir(parents=True, exist_ok=True)
        self.sync_dir.mkdir(parents=True, exist_ok=True)
        self.datasites_dir.mkdir(parents=True, exist_ok=True)
        self.plugins_dir.mkdir(parents=True, exist_ok=True)
        self.apps_dir.mkdir(parents=True, exist_ok=True)



from pathlib import Path

from fastapi import Request
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self, Union


class ServerSettings(BaseSettings):
    """
    Reads the server settings from the environment variables, using the prefix SYFTBOX_.

    example:
    `export SYFTBOX_DATA_FOLDER=data/data_folder`
    will set the server_settings.data_folder to `data/data_folder`

    see: https://docs.pydantic.dev/latest/concepts/pydantic_settings/#parsing-environment-variable-values
    """

    model_config = SettingsConfigDict(env_prefix="SYFTBOX_")

    data_folder: Path = Path("data").resolve()

    @property
    def folders(self) -> list[Path]:
        return [self.data_folder, self.snapshot_folder]

    @property
    def snapshot_folder(self) -> Path:
        return self.data_folder / "snapshot"

    @property
    def user_file_path(self) -> Path:
        return self.data_folder / "users.json"

    @classmethod
    def from_data_folder(cls, data_folder: Union[Path, str]) -> Self:
        data_folder = Path(data_folder)
        return cls(
            data_folder=data_folder,
        )

    @property
    def file_db_path(self) -> Path:
        return self.data_folder / "file.db"

    def read(self, path: Path) -> bytes:
        with open(self.snapshot_folder / path, "rb") as f:
            return f.read()


def get_server_settings(request: Request) -> ServerSettings:
    return request.state.server_settings



import json
import os

from syftbox.lib import Client

client_config = Client.load()

input_folder = (
    f"{client_config.sync_folder}/{client_config.email}/app_pipelines/adder/inputs/"
)
output_folder = (
    f"{client_config.sync_folder}/{client_config.email}/app_pipelines/adder/done/"
)
os.makedirs(input_folder, exist_ok=True)
os.makedirs(output_folder, exist_ok=True)

input_file_path = f"{input_folder}data.json"
output_file_path = f"{output_folder}data.json"

if os.path.exists(input_file_path):
    with open(input_file_path, "r") as f:
        data = json.load(f)

    data["datum"] += 1

    with open(output_file_path, "w") as f:
        json.dump(data, f)

    os.remove(input_file_path)
else:
    print(f"Input file {input_file_path} does not exist.")



import time
from collections.abc import Generator
from functools import partial
from pathlib import Path

import pytest
from fastapi.testclient import TestClient

from syftbox.client.plugins.create_datasite import run as run_create_datasite_plugin
from syftbox.client.plugins.init import run as run_init_plugin
from syftbox.lib.lib import ClientConfig, SharedState, perm_file_path
from syftbox.server.server import app as server_app
from syftbox.server.server import lifespan as server_lifespan
from syftbox.server.settings import ServerSettings


def wait_for_datasite_setup(client_config: ClientConfig, timeout=5):
    print("waiting for datasite setup...")

    perm_file = perm_file_path(str(client_config.datasite_path))

    t0 = time.time()
    while time.time() - t0 < timeout:
        perm_file_exists = Path(perm_file).exists()
        is_registered = client_config.is_registered
        if perm_file_exists and is_registered:
            print("Datasite setup complete")
            return
        time.sleep(1)

    raise TimeoutError("Datasite setup took too long")


def setup_datasite(tmp_path: Path, server_client: TestClient, email: str) -> ClientConfig:
    client_path = tmp_path / email
    client_path.unlink(missing_ok=True)
    client_path.mkdir(parents=True)

    client_config = ClientConfig(
        config_path=str(client_path / "client_config.json"),
        sync_folder=str(client_path / "sync"),
        email=email,
        server_url=str(server_client.base_url),
        autorun_plugins=[],
    )

    client_config._server_client = server_client

    shared_state = SharedState(client_config=client_config)
    run_init_plugin(shared_state)
    run_create_datasite_plugin(shared_state)
    wait_for_datasite_setup(client_config)
    return client_config


@pytest.fixture(scope="function")
def datasite_1(tmp_path: Path, server_client: TestClient) -> ClientConfig:
    email = "user_1@openmined.org"
    return setup_datasite(tmp_path, server_client, email)


@pytest.fixture(scope="function")
def datasite_2(tmp_path: Path, server_client: TestClient) -> ClientConfig:
    email = "user_2@openmined.org"
    return setup_datasite(tmp_path, server_client, email)


@pytest.fixture(scope="function")
def server_client(tmp_path: Path) -> Generator[TestClient, None, None]:
    print("Using test dir", tmp_path)
    path = tmp_path / "server"
    path.mkdir()

    settings = ServerSettings.from_data_folder(path)
    lifespan_with_settings = partial(server_lifespan, settings=settings)
    server_app.router.lifespan_context = lifespan_with_settings

    with TestClient(server_app) as client:
        yield client



import json
import os
import shutil
import subprocess
import threading
import time
from datetime import datetime
from types import SimpleNamespace

from croniter import croniter
from loguru import logger
from typing_extensions import Any, Optional, Union

from syftbox.lib import (
    SyftPermission,
    get_file_hash,
    perm_file_path,
)


def find_and_run_script(task_path, extra_args):
    script_path = os.path.join(task_path, "run.sh")
    env = os.environ.copy()  # Copy the current environment

    # Check if the script exists
    if os.path.isfile(script_path):
        # Set execution bit (+x)
        os.chmod(script_path, os.stat(script_path).st_mode | 0o111)

        # Check if the script has a shebang
        with open(script_path, "r") as script_file:
            first_line = script_file.readline().strip()
            has_shebang = first_line.startswith("#!")

        # Prepare the command based on whether there's a shebang or not
        command = (
            [script_path] + extra_args
            if has_shebang
            else ["/bin/bash", script_path] + extra_args
        )

        try:
            result = subprocess.run(
                command,
                cwd=task_path,
                check=True,
                capture_output=True,
                text=True,
                env=env,
            )

            # logger.info("✅ Script run.sh executed successfully.")
            return result
        except Exception as e:
            logger.info("Error running shell script", e)
    else:
        raise FileNotFoundError(f"run.sh not found in {task_path}")


DEFAULT_SCHEDULE = 10000
DESCRIPTION = "Runs Apps"
RUNNING_APPS = {}
DEFAULT_APPS_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "..", "..", "default_apps")
)


def copy_default_apps(apps_path):
    if not os.path.exists(DEFAULT_APPS_PATH):
        logger.info(f"Default apps directory not found: {DEFAULT_APPS_PATH}")
        return

    for app in os.listdir(DEFAULT_APPS_PATH):
        src_app_path = os.path.join(DEFAULT_APPS_PATH, app)
        dst_app_path = os.path.join(apps_path, app)

        if os.path.isdir(src_app_path):
            if os.path.exists(dst_app_path):
                logger.info(f"App already installed at: {dst_app_path}")
                # shutil.rmtree(dst_app_path)
            else:
                shutil.copytree(src_app_path, dst_app_path)
            logger.info(f"Copied default app: {app}")


def dict_to_namespace(data) -> Union[SimpleNamespace, list, Any]:
    if isinstance(data, dict):
        return SimpleNamespace(
            **{key: dict_to_namespace(value) for key, value in data.items()}
        )
    elif isinstance(data, list):
        return [dict_to_namespace(item) for item in data]
    else:
        return data


def load_config(path: str) -> Optional[SimpleNamespace]:
    try:
        with open(path, "r") as f:
            data = json.load(f)
        return dict_to_namespace(data)
    except Exception:
        return None


def run_apps(client_config):
    # create the directory
    apps_path = client_config.sync_folder + "/" + "apps"
    os.makedirs(apps_path, exist_ok=True)

    # Copy default apps if they don't exist
    copy_default_apps(apps_path)

    # add the first perm file
    file_path = perm_file_path(apps_path)
    if os.path.exists(file_path):
        perm_file = SyftPermission.load(file_path)
    else:
        logger.info(f"> {client_config.email} Creating Apps Permfile")
        try:
            perm_file = SyftPermission.datasite_default(client_config.email)
            perm_file.save(file_path)
        except Exception as e:
            logger.error("Failed to create perm file")
            logger.exception(e)

    apps = os.listdir(apps_path)
    for app in apps:
        app_path = os.path.abspath(apps_path + "/" + app)
        if os.path.isdir(app_path):
            app_config = load_config(app_path + "/" + "config.json")
            if app_config is None:
                run_app(client_config, app_path)
            elif RUNNING_APPS.get(app, None) is None:
                logger.info("⏱  Scheduling a  new app run.")
                thread = threading.Thread(
                    target=run_custom_app_config,
                    args=(client_config, app_config, app_path),
                )
                thread.start()
                RUNNING_APPS[app] = thread


def output_published(app_output, published_output) -> bool:
    return (
        os.path.exists(app_output)
        and os.path.exists(published_output)
        and get_file_hash(app_output) == get_file_hash(published_output)
    )


def run_custom_app_config(client_config, app_config, path):
    env = os.environ.copy()
    app_name = os.path.basename(path)

    # Update environment with any custom variables in app_config
    app_envs = getattr(app_config.app, "env", {})
    if not isinstance(app_envs, dict):
        app_envs = vars(app_envs)
    env.update(app_envs)

    # Retrieve the cron-style schedule from app_config
    cron_iter = None
    interval = None
    cron_schedule = getattr(app_config.app.run, "schedule", None)
    if cron_schedule is not None:
        base_time = datetime.now()
        cron_iter = croniter(cron_schedule, base_time)
    elif getattr(app_config.app.run, "interval", None) is not None:
        interval = app_config.app.run.interval
    else:
        raise Exception(
            "There's no schedule configuration. Please add schedule or interval in your app config.json"
        )

    while True:
        current_time = datetime.now()
        logger.info(
            f"👟 Running {app_name} at scheduled time {current_time.strftime('%Y-%m-%d %H:%M:%S')}"
        )
        try:
            result = subprocess.run(
                app_config.app.run.command,
                cwd=path,
                check=True,
                capture_output=True,
                text=True,
                env=env,
            )
            logger.info(result.stdout)
            logger.error(result.stderr)
        except subprocess.CalledProcessError as e:
            logger.error(f"Error running {app_name}: {e}")
            logger.error(e.stderr)

        if cron_iter is not None:
            # Schedule the next exection
            next_execution = cron_iter.get_next(datetime)
            time_to_wait = int((next_execution - current_time).total_seconds())
            logger.info(
                f"⏲ Waiting for scheduled time. Current time: {current_time.strftime('%Y-%m-%d %H:%M:%S')}, Next execution: {next_execution.strftime('%Y-%m-%d %H:%M:%S')}"
            )
        else:
            time_to_wait = int(interval)
        time.sleep(time_to_wait)


def run_app(client_config, path):
    app_name = os.path.basename(path)

    extra_args = []
    try:
        logger.info(f"👟 Running {app_name} app", end="")
        result = find_and_run_script(path, extra_args)
        if hasattr(result, "returncode"):
            if "Already generated" not in str(result.stdout):
                logger.info("\n")
                logger.info(result.stdout)
            else:
                logger.info(" - no change")
            exit_code = result.returncode
            if exit_code != 0:
                logger.info(f"Error running: {app_name}", result.stdout, result.stderr)
    except Exception as e:
        logger.info(f"Failed to run. {e}")


def run(shared_state):
    # logger.info("> Running Apps")
    client_config = shared_state.client_config
    run_apps(client_config)



import threading
from dataclasses import dataclass
from queue import PriorityQueue
from typing import Optional

from syftbox.client.plugins.sync.sync import FileChangeInfo


@dataclass(order=True)
class SyncQueueItem:
    priority: int
    data: FileChangeInfo


class SyncQueue:
    """
    A thread-safe priority queue that supports deduplication based on the data field.

    Adding an item to the queue that already exists will be ignored, even if the priority is different.
    """

    def __init__(self, maxsize: int = 0):
        self.queue = PriorityQueue(maxsize=maxsize)
        self.dedupe_set = set()

        # Lock is required to make put/get atomic when threading is used
        self.lock = threading.Lock()

    def put(self, item: SyncQueueItem, block: bool = False, timeout: Optional[float] = None) -> None:
        with self.lock:
            if item.data not in self.dedupe_set:
                self.queue.put(item, block=block, timeout=timeout)
                self.dedupe_set.add(item.data)

    def get(self, block: bool = False, timeout: Optional[float] = None) -> SyncQueueItem:
        with self.lock:
            item: SyncQueueItem = self.queue.get(block=block, timeout=timeout)
            self.dedupe_set.discard(item.data)
            return item

    def empty(self) -> bool:
        return self.queue.empty()



import logging
import os

from syftbox.lib import SyftPermission, perm_file_path
from syftbox.lib.ignore import create_default_ignore_file
from syftbox.lib.lib import Client

logger = logging.getLogger(__name__)

DEFAULT_SCHEDULE = 10000
DESCRIPTION = "Creates a datasite with a permfile"


def claim_datasite(client_config: Client):
    # create the directory
    os.makedirs(client_config.datasite_path, exist_ok=True)

    # create syftignore
    create_default_ignore_file(client_config)

    # add the first perm file
    file_path = perm_file_path(str(client_config.datasite_path))
    if os.path.exists(file_path):
        perm_file = SyftPermission.load(file_path)
    else:
        logger.info(f"> {client_config.email} Creating Datasite + Permfile")
        try:
            perm_file = SyftPermission.datasite_default(client_config.email)
            perm_file.save(file_path)
        except Exception as e:
            logger.error("Failed to create perm file")
            logger.exception(e)

    public_path = str(client_config.datasite_path) + "/" + "public"
    os.makedirs(public_path, exist_ok=True)
    public_file_path = perm_file_path(public_path)
    if os.path.exists(public_file_path):
        public_perm_file = SyftPermission.load(public_file_path)
    else:
        logger.info(f"> {client_config.email} Creating Public Permfile")
        try:
            public_perm_file = SyftPermission.mine_with_public_read(client_config.email)
            public_perm_file.save(public_file_path)
        except Exception as e:
            logger.error("Failed to create perm file")
            logger.exception(e)


def run(shared_state):
    client_config = shared_state.client_config
    claim_datasite(client_config)



import time
from threading import Thread

from loguru import logger

from syftbox.client.plugins.sync.consumer import SyncConsumer
from syftbox.client.plugins.sync.endpoints import list_datasites
from syftbox.client.plugins.sync.queue import SyncQueue, SyncQueueItem
from syftbox.client.plugins.sync.sync import DatasiteState, FileChangeInfo
from syftbox.lib import Client


class SyncManager:
    def __init__(self, client: Client):
        self.client = client
        self.queue = SyncQueue()
        self.consumer = SyncConsumer(client=self.client, queue=self.queue)

    def start(self):
        def _start(manager: SyncManager):
            while True:
                manager.run_single_thread()
                time.sleep(1)

        t = Thread(target=_start, args=[self])
        t.start()

    def setup(self):
        self.change_log_folder.mkdir(exist_ok=True)

    def enqueue(self, change: FileChangeInfo) -> None:
        self.queue.put(SyncQueueItem(priority=change.get_priority(), data=change))

    def get_datasites(self) -> list[DatasiteState]:
        try:
            datasite_names = list_datasites(self.client.server_client)
        except Exception as e:
            logger.error(f"Failed to retrieve datasites from server, only syncing own datasite. Reason: {e}")
            datasite_names = []

        # Ensure we are always syncing own datasite
        if self.client.email not in datasite_names:
            datasite_names.append(self.client.email)

        datasites = [DatasiteState(client=self.client, email=email) for email in datasite_names]
        return datasites

    def enqueue_datasite_changes(self, datasite: DatasiteState):
        try:
            permission_changes, file_changes = datasite.get_out_of_sync_files()
            logger.debug(
                f"Enqueuing {len(permission_changes)} permissions and {len(file_changes)} files for {datasite.email}"
            )
        except Exception as e:
            logger.error(f"Failed to get out of sync files for {datasite.email}. Reason: {e}")
            permission_changes, file_changes = [], []

        for change in permission_changes + file_changes:
            self.enqueue(change)

    def run_single_thread(self):
        # NOTE first implementation will be unthreaded and just loop through all datasites
        self.datasites = self.get_datasites()
        logger.debug(f"Syncing {len(self.datasites)} datasites: {[datasite.email for datasite in self.datasites]}")

        for datasite in self.datasites:
            self.enqueue_datasite_changes(datasite)
            self.consumer.consume_all()



import os
import shutil
import sqlite3
import tempfile
from pathlib import Path
from typing import Optional

from loguru import logger

from syftbox.server.settings import ServerSettings
from syftbox.server.sync.models import FileMetadata


# @contextlib.contextmanager
def get_db(path: str):
    conn = sqlite3.connect(path, check_same_thread=False)

    with conn:
        conn.execute("PRAGMA cache_size=10000;")
        conn.execute("PRAGMA synchronous=OFF;")
        conn.execute("PRAGMA journal_mode=WAL;")
        conn.execute("PRAGMA busy_timeout=5000;")
        # Create the table if it doesn't exist
        conn.execute("""
        CREATE TABLE IF NOT EXISTS file_metadata (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            path TEXT NOT NULL UNIQUE,
            hash TEXT NOT NULL,
            signature TEXT NOT NULL,
            file_size INTEGER NOT NULL,
            last_modified TEXT NOT NULL        )
        """)
    return conn


def save_file_metadata(conn: sqlite3.Connection, metadata: FileMetadata):
    # Insert the metadata into the database or update if a conflict on 'path' occurs
    conn.execute(
        """
    INSERT INTO file_metadata (path, hash, signature, file_size, last_modified)
    VALUES (?, ?, ?, ?, ?)
    ON CONFLICT(path) DO UPDATE SET
        hash = excluded.hash,
        signature = excluded.signature,
        file_size = excluded.file_size,
        last_modified = excluded.last_modified
    """,
        (
            str(metadata.path),
            metadata.hash,
            metadata.signature,
            metadata.file_size,
            metadata.last_modified.isoformat(),
        ),
    )


def delete_file_metadata(conn: sqlite3.Connection, path: str):
    cur = conn.execute("DELETE FROM file_metadata WHERE path = ?", (path,))
    # get number of changes
    if cur.rowcount != 1:
        raise ValueError(f"Failed to delete metadata for {path}.")


def get_all_metadata(conn: sqlite3.Connection, path_like: Optional[str] = None) -> list[FileMetadata]:
    query = "SELECT * FROM file_metadata"
    params = ()

    if path_like:
        query += " WHERE path LIKE ?"
        params = (path_like,)

    cursor = conn.execute(query, params)
    # would be nice to paginate
    return [
        FileMetadata(
            path=row[1],
            hash=row[2],
            signature=row[3],
            file_size=row[4],
            last_modified=row[5],
        )
        for row in cursor
    ]


def get_all_datasites(conn: sqlite3.Connection) -> list[str]:
    # INSTR(path, '/'): Finds the position of the first slash in the path.
    cursor = conn.execute(
        """SELECT DISTINCT SUBSTR(path, 1, INSTR(path, '/') - 1) AS root_folder
        FROM file_metadata;
        """
    )
    return [row[0] for row in cursor]


def move_with_transaction(
    conn: sqlite3.Connection, *, origin_path: Path, metadata: FileMetadata, server_settings: ServerSettings
):
    """The file system and database do not share transactions,
    so this operation is not atomic.
    Ideally, files (blobs) should be immutable,
    and the path should update to a new location
    whenever there is a change to the file contents.
    """

    # backup the original file
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_path = temp_file.name

    shutil.copy(origin_path, temp_path)

    cursor = conn.cursor()
    try:
        # Update database entry
        from_path = metadata.path
        relative_path = origin_path.relative_to(server_settings.snapshot_folder)
        metadata.path = relative_path
        save_file_metadata(conn, metadata)

        conn.commit()

        # WARNING: between the move and the commit
        # the database will be in an inconsistent state

        shutil.move(from_path, origin_path)

    except sqlite3.IntegrityError as e:
        # Rollback the transaction in case of error
        conn.rollback()
        logger.error(f"Failed to update metadata for {metadata.path}. Rolled back.")

        # raise the original error
        raise e
    finally:
        # Clean up
        cursor.close()
        conn.close()

        # Delete the temp file if it exists
        if os.path.exists(temp_path):
            os.remove(temp_path)



from threading import Event

from loguru import logger

stop_event = Event()


def register(client_config):
    response = client_config.server_client.post(
        "/register",
        json={"email": client_config.email},
    )

    j = response.json()
    if "token" in j:
        client_config.token = j["token"]
        client_config.save()

    return response.status_code == 200


def run(shared_state):
    if not stop_event.is_set():
        if not shared_state.client_config.token:
            register(shared_state.client_config)
            logger.info("> Register Complete")



from pathlib import Path
from typing import Optional

import pathspec
from loguru import logger

from syftbox.lib import Client

IGNORE_FILENAME = "_.syftignore"

DEFAULT_IGNORE = """
# Syft
/_.syftignore
/.syft*
/apps
/staging
/syft_changelog

# Python
.ipynb_checkpoints/
__pycache__/
*.py[cod]
.venv/

# OS-specific
.DS_Store
Icon

# IDE/Editor-specific
*.swp
*.swo
.vscode/
.idea/
*.iml

# General excludes
*.tmp

# excluded datasites
# example:
# /user_to_exclude@example.com/
"""


def create_default_ignore_file(client: Client) -> None:
    ignore_file = Path(client.sync_folder) / IGNORE_FILENAME
    if not ignore_file.is_file():
        logger.info(f"Creating default ignore file: {ignore_file}")
        ignore_file.write_text(DEFAULT_IGNORE)


def get_ignore_rules(client: Client) -> Optional[pathspec.PathSpec]:
    ignore_file = Path(client.sync_folder) / IGNORE_FILENAME
    if ignore_file.is_file():
        with open(ignore_file) as f:
            lines = f.readlines()
        return pathspec.PathSpec.from_lines("gitwildmatch", lines)
    return None


def filter_ignored_paths(client: Client, paths: list[Path]) -> list[Path]:
    ignore_rules = get_ignore_rules(client)
    if ignore_rules is None:
        return paths

    filtered_paths = []
    for path in paths:
        if not ignore_rules.match_file(path):
            filtered_paths.append(path)

    return filtered_paths



import base64
import sqlite3
import tempfile
from pathlib import Path

import py_fast_rsync
from fastapi import APIRouter, Depends, Header, HTTPException, Request, UploadFile
from fastapi.responses import FileResponse, JSONResponse

from syftbox.lib.lib import PermissionTree, filter_metadata
from syftbox.server.settings import ServerSettings, get_server_settings
from syftbox.server.sync.db import (
    delete_file_metadata,
    get_all_datasites,
    get_all_metadata,
    get_db,
    move_with_transaction,
    save_file_metadata,
)
from syftbox.server.sync.hash import hash_file

from .models import (
    ApplyDiffRequest,
    ApplyDiffResponse,
    DiffRequest,
    DiffResponse,
    FileMetadata,
    FileMetadataRequest,
    FileRequest,
)


def get_db_connection(request: Request):
    conn = get_db(request.state.server_settings.file_db_path)
    yield conn
    conn.close()


def get_file_metadata(
    req: FileMetadataRequest,
    conn=Depends(get_db_connection),
) -> list[FileMetadata]:
    # TODO check permissions

    return get_all_metadata(conn, path_like=req.path_like)


router = APIRouter(prefix="/sync", tags=["sync"])


@router.post("/get_diff", response_model=DiffResponse)
def get_diff(
    req: DiffRequest,
    conn: sqlite3.Connection = Depends(get_db_connection),
    server_settings: ServerSettings = Depends(get_server_settings),
) -> DiffResponse:
    metadata_list = get_all_metadata(conn, path_like=f"{req.path}")
    if len(metadata_list) == 0:
        raise HTTPException(status_code=404, detail="path not found")
    elif len(metadata_list) > 1:
        raise HTTPException(status_code=400, detail="too many files to get diff")

    metadata = metadata_list[0]
    abs_path = server_settings.snapshot_folder / metadata.path
    with open(abs_path, "rb") as f:
        data = f.read()

    diff = py_fast_rsync.diff(req.signature_bytes, data)
    diff_bytes = base64.b85encode(diff).decode("utf-8")
    return DiffResponse(
        path=metadata.path.as_posix(),
        diff=diff_bytes,
        hash=metadata.hash,
    )


@router.post("/dir_state", response_model=list[FileMetadata])
def dir_state(
    dir: Path,
    conn: sqlite3.Connection = Depends(get_db_connection),
    server_settings: ServerSettings = Depends(get_server_settings),
    email: str = Header(),
) -> list[FileMetadata]:
    if dir.is_absolute():
        raise HTTPException(status_code=400, detail="dir must be relative")

    metadata = get_all_metadata(conn, path_like=f"{dir.as_posix()}%")
    full_path = server_settings.snapshot_folder / dir
    # get the top level perm file
    try:
        perm_tree = PermissionTree.from_path(full_path, raise_on_corrupted_files=True)
    except Exception as e:
        print(f"Failed to parse permission tree: {dir}")
        raise e

    # filter the read state for this user by the perm tree
    filtered_metadata = filter_metadata(email, metadata, perm_tree, server_settings.snapshot_folder)
    return filtered_metadata


@router.post("/get_metadata", response_model=list[FileMetadata])
def get_metadata(
    metadata: list[FileMetadata] = Depends(get_file_metadata),
) -> list[FileMetadata]:
    return metadata


@router.post("/apply_diff", response_model=ApplyDiffResponse)
def apply_diffs(
    req: ApplyDiffRequest,
    conn: sqlite3.Connection = Depends(get_db_connection),
    server_settings: ServerSettings = Depends(get_server_settings),
) -> ApplyDiffResponse:
    metadata_list = get_all_metadata(conn, path_like=f"{req.path}")

    if len(metadata_list) == 0:
        raise HTTPException(status_code=404, detail="path not found")
    elif len(metadata_list) > 1:
        raise HTTPException(status_code=400, detail="found too many files to apply diff")

    metadata = metadata_list[0]

    abs_path = server_settings.snapshot_folder / metadata.path
    with open(abs_path, "rb") as f:
        data = f.read()
    result = py_fast_rsync.apply(data, req.diff_bytes)

    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_file.write(result)
        temp_path = Path(temp_file.name)

    new_metadata = hash_file(temp_path)

    if new_metadata.hash != req.expected_hash:
        raise HTTPException(status_code=400, detail="expected_hash mismatch")

    # move temp path to real path and update db
    move_with_transaction(conn, metadata=new_metadata, origin_path=abs_path, server_settings=server_settings)

    return ApplyDiffResponse(path=req.path, current_hash=new_metadata.hash, previous_hash=metadata.hash)


@router.post("/delete", response_class=JSONResponse)
def delete_file(
    req: FileRequest,
    conn: sqlite3.Connection = Depends(get_db_connection),
    server_settings: ServerSettings = Depends(get_server_settings),
) -> JSONResponse:
    metadata_list = get_all_metadata(conn, path_like=f"{req.path}")
    if len(metadata_list) == 0:
        raise HTTPException(status_code=404, detail="path not found")
    elif len(metadata_list) > 1:
        raise HTTPException(status_code=400, detail="too many files to delete")

    metadata = metadata_list[0]

    try:
        delete_file_metadata(conn, metadata.path.as_posix())
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))
    conn.commit()
    abs_path = server_settings.snapshot_folder / metadata.path
    Path(abs_path).unlink(missing_ok=True)
    return JSONResponse(content={"status": "success"})


@router.post("/create", response_class=JSONResponse)
def create_file(
    file: UploadFile,
    conn: sqlite3.Connection = Depends(get_db_connection),
    server_settings: ServerSettings = Depends(get_server_settings),
) -> JSONResponse:
    #
    relative_path = Path(file.filename)
    abs_path = server_settings.snapshot_folder / relative_path
    abs_path.parent.mkdir(exist_ok=True, parents=True)
    with open(abs_path, "wb") as f:
        # better to use async aiosqlite
        f.write(file.file.read())

    cursor = conn.cursor()
    metadata = get_all_metadata(cursor, path_like=f"{file.filename}")
    if len(metadata) > 0:
        raise HTTPException(status_code=400, detail="file already exists")
    metadata = hash_file(abs_path, root_dir=server_settings.snapshot_folder)
    save_file_metadata(cursor, metadata)
    conn.commit()
    cursor.close()

    return JSONResponse(content={"status": "success"})


@router.post("/download", response_class=FileResponse)
def download_file(
    req: FileRequest,
    conn: sqlite3.Connection = Depends(get_db_connection),
    server_settings: ServerSettings = Depends(get_server_settings),
) -> FileResponse:
    metadata_list = get_all_metadata(conn, path_like=f"{req.path}")
    if len(metadata_list) == 0:
        raise HTTPException(status_code=404, detail="path not found")
    elif len(metadata_list) > 1:
        raise HTTPException(status_code=400, detail="too many files to download")

    metadata = metadata_list[0]
    abs_path = server_settings.snapshot_folder / metadata.path
    if not Path(abs_path).exists():
        raise HTTPException(status_code=404, detail="file not found")
    return FileResponse(abs_path)


@router.post("/datasites", response_model=list[str])
def get_datasites(conn: sqlite3.Connection = Depends(get_db_connection)) -> list[str]:
    return get_all_datasites(conn)



import base64
import hashlib
import re
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from pathlib import Path
from typing import Optional, Union

from loguru import logger
from py_fast_rsync import signature

from syftbox.server.sync.models import FileMetadata


def hash_file(file_path: Path, root_dir: Optional[Path] = None) -> Optional[FileMetadata]:
    # ignore files larger then 100MB
    try:
        if file_path.stat().st_size > 100_000_000:
            logger.warning("File too large: %s", file_path)
            return str(file_path), None

        with open(file_path, "rb") as f:
            # not ideal for large files
            # but py_fast_rsync does not support files yet.
            # TODO: add support for streaming hashing
            data = f.read()

        if root_dir is None:
            path = file_path
        else:
            path = file_path.relative_to(root_dir)
        return FileMetadata(
            path=path,
            hash=hashlib.sha256(data).hexdigest(),
            signature=base64.b85encode(signature.calculate(data)),
            file_size=len(data),
            last_modified=file_path.stat().st_mtime,
        )
    except Exception:
        logger.error(f"Failed to hash file {file_path}")
        return None


def hash_files_parallel(files: list[Path], root_dir: Path) -> list[FileMetadata]:
    with ProcessPoolExecutor() as executor:
        results = list(executor.map(partial(hash_file, root_dir=root_dir), files))
    return [r for r in results if r is not None]


def hash_files(files: list[Path], root_dir: Path) -> list[FileMetadata]:
    result = [hash_file(file, root_dir) for file in files]
    return [r for r in result if r is not None]


def hash_dir(dir: Path, root_dir: Path) -> list[FileMetadata]:
    """
    hash all files in dir recursively, return a list of FileMetadata.

    ignore_folders should be relative to root_dir.
    returned Paths are relative to root_dir.
    """
    files = collect_files(dir)
    return hash_files(files, root_dir)


def collect_files(dir: Union[Path, str], pattern: Union[str, re.Pattern, None] = None) -> list[Path]:
    """Recursively collect files in a directory

    Examples:
        >>> # list all .syftperm files
        >>> collect_files(snapshot_dir, r".*/.syftperm")

        >>> # list all files in a directory info
        >>> collect_files(snapshot_dir, r".*")


    """
    dir = Path(dir)
    if not dir.is_dir():
        return []
    files = []

    # Compile the regex pattern if it's a string
    if isinstance(pattern, str):
        pattern = re.compile(pattern)

    for entry in dir.iterdir():
        if entry.is_file():
            if pattern is None or pattern.match(entry.as_posix()):
                files.append(entry)
        elif entry.is_dir():
            files.extend(collect_files(entry, pattern))

    return files



import os
import shutil
import subprocess
from pathlib import Path

default_apps = [
    "https://github.com/OpenMined/ring",
    "https://github.com/OpenMined/github_app_updater",
    "https://github.com/OpenMined/logged_in",
]


def clone_apps():
    global default_apps
    # Iterate over the list and clone each repository
    for url in default_apps:
        subprocess.run(["git", "clone", url])


if __name__ == "__main__":
    current_directory = Path(os.getcwd())

    apps_directory = current_directory.parent
    os.chdir(apps_directory)
    clone_apps()
    shutil.rmtree(current_directory)



import os
import shutil

import pytest

from syftbox.app.install import clone_repository, sanitize_git_path


def test_valid_git_path():
    path = "Example/Repository"
    output_path = sanitize_git_path(path)
    assert path == output_path


def test_valid_git_url():
    path = "Example/Repository"
    http_url = f"http://github.com/{path}"
    output_path = sanitize_git_path(http_url)
    assert path == output_path

    https_url = f"https://github.com/{path}"
    output_path = sanitize_git_path(https_url)
    assert path == output_path


def test_invalid_git_path():
    path = "..Example/../Repository"
    with pytest.raises(ValueError) as excpt:
        _ = sanitize_git_path(path)
        assert excpt.value == "Invalid Git repository path format."


def test_second_invalid_git_path():
    path = "http://example.com"
    with pytest.raises(ValueError) as excpt:
        _ = sanitize_git_path(path)
        assert excpt.value == "Invalid Git repository path format."


def test_clone_valid_repository():
    path = "OpenMined/logged_in"
    temp_path = clone_repository(path)
    assert os.path.exists(temp_path)
    shutil.rmtree(temp_path)


def test_clone_repository_to_an_existent_path():
    # First call will make the repository path exist
    path = "OpenMined/logged_in"
    temp_path = clone_repository(path)
    assert os.path.exists(temp_path)

    # Second call must clone it again without any exception (replaces the old one).
    temp_path = clone_repository(path)
    shutil.rmtree(temp_path)


def test_clone_invalid_repository():
    path = "InvalidUser/InvalidRepo"
    with pytest.raises(ValueError) as excpt:
        _ = clone_repository(path)
        assert excpt.value == "The provided repository path doesn't seems to be accessible. Please check it out."



import os
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Optional, Union

from loguru import logger

from syftbox.client.plugins.sync.endpoints import get_remote_state
from syftbox.lib import Client, DirState
from syftbox.lib.ignore import filter_ignored_paths
from syftbox.server.models import SyftBaseModel
from syftbox.server.sync.hash import hash_dir
from syftbox.server.sync.models import FileMetadata


def is_permission_file(path: Union[Path, str], check_exists: bool = False) -> bool:
    path = Path(path)
    if check_exists and not path.is_file():
        return False

    return path.name == "_.syftperm"


class SyncSide(str, Enum):
    LOCAL = "local"
    REMOTE = "remote"


class FileChangeInfo(SyftBaseModel, frozen=True):
    path: Path
    side_last_modified: SyncSide
    date_last_modified: datetime
    file_size: int = 1

    def get_priority(self) -> int:
        if is_permission_file(self.path):
            return 0
        else:
            return max(1, self.file_size)

    def __lt__(self, other: "FileChangeInfo") -> bool:
        return self.path < other.path


class DatasiteState:
    def __init__(self, client: Client, email: str):
        """
        NOTE DatasiteState is not threadsafe, this should be handled by the caller
        """
        self.client = client
        self.email = email

    @property
    def path(self) -> Path:
        p = Path(self.client.sync_folder) / self.email
        return p.expanduser().resolve()

    def get_current_local_state(self) -> list[FileMetadata]:
        return hash_dir(self.path, root_dir=self.client.sync_folder)

    def get_remote_state(self) -> list[FileMetadata]:
        return get_remote_state(self.client.server_client, email=self.client.email, path=Path(self.email))

    def get_out_of_sync_files(
        self,
    ) -> tuple[list[FileChangeInfo], list[FileChangeInfo]]:
        """
        calculate the files that are out of sync

        NOTE: we are not handling local permissions here,
        they will be handled by the server and consumer
        TODO: we are not handling empty folders
        """
        try:
            local_state = self.get_current_local_state()
        except Exception:
            logger.error(f"Failed to get local state for {self.email}")
            return [], []

        try:
            remote_state = self.get_remote_state()
        except Exception:
            logger.error(f"Failed to get remote state from server {self.email}")
            return [], []

        local_state_dict = {file.path: file for file in local_state}
        remote_state_dict = {file.path: file for file in remote_state}
        all_files = set(local_state_dict.keys()) | set(remote_state_dict.keys())
        all_files_filtered = filter_ignored_paths(client=self.client, paths=list(all_files))

        all_changes = []

        for afile in all_files_filtered:
            local_info = local_state_dict.get(afile)
            remote_info = remote_state_dict.get(afile)

            try:
                change_info = compare_fileinfo(afile, local_info, remote_info)
            except Exception:
                logger.exception(f"Failed to compare file {afile}")
                continue

            if change_info is not None:
                all_changes.append(change_info)

        # TODO implement ignore rules
        # ignore_rules = get_ignore_rules(local_state)
        # filtered_changes = filter_ignored_changes(all_changes, ignore_rules)

        permission_changes, file_changes = split_permissions(all_changes)
        # TODO debounce changes
        # filtered_changes = filter_recent_local_changes(filtered_changes)

        return permission_changes, file_changes


def split_permissions(
    changes: list[FileChangeInfo],
) -> tuple[list[FileChangeInfo], list[FileChangeInfo]]:
    permissions = []
    files = []
    for change in changes:
        if is_permission_file(change.path):
            permissions.append(change)
        else:
            files.append(change)
    return permissions, files


def compare_fileinfo(
    path: Path,
    local_info: Optional[FileMetadata],
    remote_info: Optional[FileMetadata],
) -> Optional[FileChangeInfo]:
    if local_info is None and remote_info is None:
        return

    if local_info is None and remote_info is not None:
        # File only exists on remote
        return FileChangeInfo(
            path=path,
            side_last_modified=SyncSide.REMOTE,
            date_last_modified=remote_info.last_modified,
            file_size=remote_info.file_size,
        )

    if remote_info is None and local_info is not None:
        # File only exists on local
        return FileChangeInfo(
            path=path,
            side_last_modified=SyncSide.LOCAL,
            date_last_modified=local_info.last_modified,
            file_size=local_info.file_size,
        )

    if local_info.hash != remote_info.hash:
        # File is different on both sides
        if local_info.last_modified > remote_info.last_modified:
            date_last_modified = local_info.last_modified
            side_last_modified = SyncSide.LOCAL
            file_size = local_info.file_size
        else:
            date_last_modified = remote_info.last_modified
            side_last_modified = SyncSide.REMOTE
            file_size = remote_info.file_size

        return FileChangeInfo(
            path=path,
            side_last_modified=side_last_modified,
            date_last_modified=date_last_modified,
            file_size=file_size,
        )


def get_ignore_rules(dir_state: DirState) -> list[str, str, str]:
    """
    TODO refactor, abs/relative paths are not handled correctly
    returns a list of ignore rules (prefix, folder, ignore_file)
    """
    # get the ignore files
    syft_ignore_files = []
    folder_path = dir_state.sync_folder + "/" + dir_state.sub_path
    for afile, file_info in dir_state.tree.items():
        full_path = folder_path + "/" + afile
        sub_folder = os.path.dirname(full_path)

        if afile.endswith(".syftignore") and os.path.isfile(full_path):
            ignore_list = []
            with open(full_path) as f:
                ignore_list = f.readlines()
            for ignore_rule in ignore_list:
                ignore_rule = ignore_rule.strip()
                rule_prefix = sub_folder + "/" + ignore_rule
                syft_ignore_files.append((rule_prefix, sub_folder, afile))

    return syft_ignore_files


def filter_ignored_changes(
    all_changes: list[FileChangeInfo], ignore_rules: list[str, str, str]
) -> list[FileChangeInfo]:
    """
    Filter out changes that are ignored by .syftignore files
    """
    filtered_changes = []
    for change in all_changes:
        keep = True
        for rule_prefix, ignore_folder, ignore_file_path in ignore_rules:
            if change.path.startswith(rule_prefix):
                keep = False
                break
        if keep:
            filtered_changes.append(change)
    return filtered_changes



from pathlib import Path
from shutil import make_archive

from loguru import logger

from syftbox.lib.lib import DEFAULT_LOGS_PATH

# Configure Loguru to write logs to a file with rotation
logger.add(
    DEFAULT_LOGS_PATH,
    rotation="100 MB",  # Rotate after the log file reaches 100 MB
    retention=2,  # Keep only the last 1 log files
    compression="zip",  # Usually, 10x reduction in file size
)


def zip_logs(output_path):
    logger.info("Compressing logs folder")
    logs_folder = Path(DEFAULT_LOGS_PATH).parent
    return make_archive(output_path, "zip", logs_folder)



import hashlib
import threading
from enum import Enum
from pathlib import Path
from typing import Optional

import py_fast_rsync
from loguru import logger
from pydantic import BaseModel, model_validator

from syftbox.client.plugins.sync.endpoints import (
    SyftServerError,
    apply_diff,
    create,
    delete,
    download,
    get_diff,
    get_metadata,
)
from syftbox.client.plugins.sync.queue import SyncQueue, SyncQueueItem
from syftbox.client.plugins.sync.sync import SyncSide
from syftbox.lib.lib import Client
from syftbox.server.sync.hash import hash_file
from syftbox.server.sync.models import FileMetadata


class SyncDecisionType(Enum):
    NOOP = 0
    CREATE = 1
    MODIFY = 2
    DELETE = 3


def update_local(client: Client, local_syncstate: FileMetadata, remote_syncstate: FileMetadata):
    diff = get_diff(client.server_client, local_syncstate.path, remote_syncstate.signature_bytes)
    abs_path = client.sync_folder / local_syncstate.path
    local_data = abs_path.read_bytes()

    new_data = py_fast_rsync.apply(local_data, diff.diff_bytes)
    new_hash = hashlib.sha256(new_data).hexdigest()

    if new_hash != diff.hash:
        # TODO handle
        raise ValueError("hash mismatch")

    # TODO implement safe write with tempfile + rename
    abs_path.parent.mkdir(parents=True, exist_ok=True)
    abs_path.write_bytes(new_data)


def update_remote(client: Client, local_syncstate: FileMetadata, remote_syncstate: FileMetadata):
    abs_path = client.sync_folder / local_syncstate.path
    local_data = abs_path.read_bytes()

    diff = py_fast_rsync.diff(remote_syncstate.signature_bytes, local_data)
    apply_diff(client.server_client, local_syncstate.path, diff, local_syncstate.hash)


def delete_local(client: Client, remote_syncstate: FileMetadata):
    abs_path = client.sync_folder / remote_syncstate.path
    abs_path.unlink()


def delete_remote(client: Client, local_syncstate: FileMetadata):
    delete(client.server_client, local_syncstate.path)


def create_local(client: Client, remote_syncstate: FileMetadata):
    abs_path = client.sync_folder / remote_syncstate.path
    content_bytes = download(client.server_client, remote_syncstate.path)
    abs_path.parent.mkdir(parents=True, exist_ok=True)
    abs_path.write_bytes(content_bytes)


def create_remote(client: Client, local_syncstate: FileMetadata):
    abs_path = client.sync_folder / local_syncstate.path
    data = abs_path.read_bytes()
    create(client.server_client, local_syncstate.path, data)


class SyncDecision(BaseModel):
    operation: SyncDecisionType
    side_to_update: SyncSide
    local_syncstate: Optional[FileMetadata]
    remote_syncstate: Optional[FileMetadata]

    def execute(self, client: Client):
        if self.operation == SyncDecisionType.NOOP:
            return

        to_local = self.side_to_update == SyncSide.LOCAL
        to_remote = self.side_to_update == SyncSide.REMOTE

        if self.operation == SyncDecisionType.CREATE and to_remote:
            create_remote(client, self.local_syncstate)
        elif self.operation == SyncDecisionType.CREATE and to_local:
            create_local(client, self.remote_syncstate)
        elif self.operation == SyncDecisionType.DELETE and to_remote:
            delete_remote(client, self.remote_syncstate)
        elif self.operation == SyncDecisionType.DELETE and to_local:
            delete_local(client, self.local_syncstate)
        elif self.operation == SyncDecisionType.MODIFY and to_remote:
            update_remote(client, self.local_syncstate, self.remote_syncstate)
        elif self.operation == SyncDecisionType.MODIFY and to_local:
            update_local(client, self.local_syncstate, self.remote_syncstate)

    @classmethod
    def noop(
        cls,
        local_syncstate: FileMetadata,
        remote_syncstate: FileMetadata,
    ):
        return cls(
            operation=SyncDecisionType.NOOP,
            side_to_update=SyncSide.LOCAL,
            local_syncstate=local_syncstate,
            remote_syncstate=remote_syncstate,
        )

    @classmethod
    def from_modified_states(
        cls,
        local_syncstate: Optional[FileMetadata],
        remote_syncstate: Optional[FileMetadata],
        side_to_update: SyncSide,
    ):
        """Asssumes at least on of the states is modified"""

        delete = (
            side_to_update == SyncSide.REMOTE
            and local_syncstate is None
            or side_to_update == SyncSide.LOCAL
            and remote_syncstate is None
        )

        create = (
            side_to_update == SyncSide.REMOTE
            and remote_syncstate is None
            or side_to_update == SyncSide.LOCAL
            and local_syncstate is None
        )

        if delete:
            operation = SyncDecisionType.DELETE
        elif create:
            operation = SyncDecisionType.CREATE
        else:
            operation = SyncDecisionType.MODIFY

        return cls(
            operation=operation,
            side_to_update=side_to_update,
            local_syncstate=local_syncstate,
            remote_syncstate=remote_syncstate,
        )


class SyncDecisionTuple(BaseModel):
    remote_decision: SyncDecision
    local_decision: SyncDecision

    @property
    def result_local_state(self):
        if self.local_decision.operation == SyncDecisionType.NOOP:
            return self.local_decision.local_syncstate
        else:
            return self.local_decision.remote_syncstate

    @classmethod
    def from_states(
        cls,
        current_local_syncstate: Optional[FileMetadata],
        previous_local_syncstate: Optional[FileMetadata],
        current_remote_syncstate: Optional[FileMetadata],
    ):
        def noop() -> SyncDecision:
            return SyncDecision.noop(
                local_syncstate=current_local_syncstate,
                remote_syncstate=current_remote_syncstate,
            )

        local_modified = current_local_syncstate != previous_local_syncstate
        remote_modified = previous_local_syncstate != current_remote_syncstate
        in_sync = current_remote_syncstate == current_local_syncstate
        conflict = local_modified and remote_modified and not in_sync

        logger.debug(
            f"local_modified: {local_modified}, remote_modified: {remote_modified}, in_sync: {in_sync}, conflict: {conflict}"
        )

        if in_sync:
            return cls(
                remote_decision=noop(),
                local_decision=noop(),
            )
        elif conflict:
            # in case of conflict we always use the server state, because it was updated earlier
            remote_decision = noop()
            # we apply the server state locally
            local_decision = SyncDecision.from_modified_states(
                local_syncstate=current_local_syncstate,
                remote_syncstate=current_remote_syncstate,
                side_to_update=SyncSide.LOCAL,
            )
            return cls(remote_decision=remote_decision, local_decision=local_decision)
        else:
            # here we can assume only one party changed
            # assert (local_modified and not server_modified) or (server_modified and not local_modified)
            if local_modified:
                return cls(
                    local_decision=noop(),
                    remote_decision=SyncDecision.from_modified_states(
                        local_syncstate=current_local_syncstate,
                        remote_syncstate=current_remote_syncstate,
                        side_to_update=SyncSide.REMOTE,
                    ),
                )
            else:
                return cls(
                    local_decision=SyncDecision.from_modified_states(
                        local_syncstate=current_local_syncstate,
                        remote_syncstate=current_remote_syncstate,
                        side_to_update=SyncSide.LOCAL,
                    ),
                    remote_decision=noop(),
                )


class LocalState(BaseModel):
    path: Path
    states: dict[Path, FileMetadata] = {}

    @model_validator(mode="after")
    def init_dir(self):
        self.path.parent.mkdir(parents=True, exist_ok=True)
        return self

    def insert(self, path: Path, state: FileMetadata):
        if state is None:
            self.states.pop(path, None)
        else:
            self.states[path] = state
        self.save()

    def save(self):
        with threading.Lock():
            self.path.write_text(self.model_dump_json())

    def load(self):
        with threading.Lock():
            if self.path.exists():
                data = self.path.read_text()
                loaded_state = self.model_validate_json(data)
                self.states = loaded_state.states


class SyncConsumer:
    def __init__(self, client: Client, queue: SyncQueue):
        self.client = client
        self.queue = queue
        self.previous_state = LocalState(path=Path(client.sync_folder) / ".syft" / "local_syncstate.json")
        self.previous_state.load()

    def consume_all(self):
        while not self.queue.empty():
            item = self.queue.get(timeout=0.1)
            try:
                self.process_filechange(item)
            except Exception:
                logger.exception(f"Failed to sync file {item.data.path}")

    def get_decisions(self, item: SyncQueueItem) -> SyncDecisionTuple:
        path = item.data.path
        current_local_syncstate: FileMetadata = self.get_current_local_syncstate(path)
        previous_local_syncstate = self.get_previous_local_syncstate(path)
        # TODO, rename to remote
        current_server_state = self.get_current_server_state(path)

        local_hash = current_local_syncstate.hash if current_local_syncstate else None
        server_hash = current_server_state.hash if current_server_state else None
        previous_local_hash = previous_local_syncstate.hash if previous_local_syncstate else None

        logger.debug(
            f"Processing {path} with local hash {local_hash}, server hash {server_hash}, previous local hash {previous_local_hash}"
        )

        return SyncDecisionTuple.from_states(current_local_syncstate, previous_local_syncstate, current_server_state)

    def process_decision(self, path: Path, decision: SyncDecisionTuple):
        decision.remote_decision.execute(self.client)
        decision.local_decision.execute(self.client)
        logger.debug(f"Saving state for {path}, {decision.result_local_state}")

        self.previous_state.insert(path=path, state=decision.result_local_state)

    def process_filechange(self, item: SyncQueueItem) -> None:
        decisions = self.get_decisions(item)
        logger.debug(
            f"Processing {item.data.path} with decisions {decisions.local_decision.operation}, {decisions.remote_decision.operation}"
        )
        self.process_decision(item.data.path, decisions)

    def get_current_local_syncstate(self, path: Path) -> Optional[FileMetadata]:
        abs_path = self.client.sync_folder / path
        if not abs_path.is_file():
            return None
        return hash_file(abs_path, root_dir=self.client.sync_folder)

    def get_previous_local_syncstate(self, path: Path) -> Optional[FileMetadata]:
        return self.previous_state.states.get(path, None)

    def get_current_server_state(self, path: Path) -> Optional[FileMetadata]:
        try:
            return get_metadata(self.client.server_client, path)
        except SyftServerError:
            return None



import json
from functools import partial
from pathlib import Path
from typing import Generator

import pytest
from fastapi.testclient import TestClient

from syftbox.server.server import app
from syftbox.server.server import app as server_app
from syftbox.server.server import lifespan as server_lifespan
from syftbox.server.settings import ServerSettings

TEST_DATASITE_NAME = "test_datasite@openmined.org"
TEST_FILE = "test_file.txt"
PERMFILE_FILE = "_.syftperm"
PERMFILE_DICT = {
    "admin": [TEST_DATASITE_NAME],
    "read": [TEST_DATASITE_NAME],
    "write": [TEST_DATASITE_NAME],
}


@pytest.fixture(scope="function")
def client(monkeypatch, tmp_path):
    """Every client gets their own snapshot folder at `tmp_path`"""
    snapshot_folder = tmp_path / "snapshot"
    settings = ServerSettings.from_data_folder(snapshot_folder)
    monkeypatch.setenv("SYFTBOX_DATA_FOLDER", str(settings.data_folder))
    monkeypatch.setenv("SYFTBOX_SNAPSHOT_FOLDER", str(settings.snapshot_folder))
    monkeypatch.setenv("SYFTBOX_USER_FILE_PATH", str(settings.user_file_path))

    datasite_name = TEST_DATASITE_NAME
    datasite = settings.snapshot_folder / datasite_name
    datasite.mkdir(parents=True)

    datafile = datasite / TEST_FILE
    datafile.touch()
    datafile.write_bytes(b"Hello, World!")

    datafile = datasite / TEST_DATASITE_NAME / TEST_FILE
    datafile.parent.mkdir(parents=True)

    datafile.touch()
    datafile.write_bytes(b"Hello, World!")

    permfile = datasite / PERMFILE_FILE
    permfile.touch()
    permfile.write_text(json.dumps(PERMFILE_DICT))

    with TestClient(app) as client:
        yield client


@pytest.fixture(scope="function")
def client_without_perms(monkeypatch, tmp_path):
    """Every client gets their own snapshot folder at `tmp_path`"""
    settings = ServerSettings.from_data_folder(tmp_path)
    monkeypatch.setenv("SYFTBOX_DATA_FOLDER", str(settings.data_folder))
    monkeypatch.setenv("SYFTBOX_SNAPSHOT_FOLDER", str(settings.snapshot_folder))
    monkeypatch.setenv("SYFTBOX_USER_FILE_PATH", str(settings.user_file_path))

    datasite_name = TEST_DATASITE_NAME
    datasite = settings.snapshot_folder / datasite_name
    datasite.mkdir(parents=True)

    datafile = datasite / TEST_FILE
    datafile.touch()
    datafile.write_bytes(b"Hello, World!")

    permfile = datasite / PERMFILE_FILE
    permfile.touch()
    permfile.write_text("")

    with TestClient(app) as client:
        yield client


@pytest.fixture(scope="function")
def server_client(tmp_path: Path) -> Generator[TestClient, None, None]:
    print("Using test dir", tmp_path)
    path = tmp_path / "server"
    path.mkdir()

    settings = ServerSettings.from_data_folder(path)
    lifespan_with_settings = partial(server_lifespan, settings=settings)
    server_app.router.lifespan_context = lifespan_with_settings

    with TestClient(server_app) as client:
        yield client



import base64
from pathlib import Path
from typing import Any

import httpx

from syftbox.server.sync.models import ApplyDiffResponse, DiffResponse, FileMetadata


class SyftServerError(Exception):
    pass


class SyftNotFound(SyftServerError):
    pass


def handle_json_response(endpoint: str, response: httpx.Response) -> Any:
    # endpoint only needed for error message
    if response.status_code == 200:
        return response.json()

    raise SyftServerError(f"[{endpoint}] call failed: {response.text}")


def list_datasites(client: httpx.Client) -> list[str]:
    response = client.post(
        "/sync/datasites",
    )

    data = handle_json_response("/sync/datasites", response)
    return data


def get_remote_state(client: httpx.Client, email: str, path: Path) -> list[FileMetadata]:
    response = client.post(
        "/sync/dir_state",
        params={
            "dir": str(path),
        },
        headers={"email": email},
    )

    response_data = handle_json_response("/dir_state", response)
    return [FileMetadata(**item) for item in response_data]


def get_metadata(client: httpx.Client, path: Path) -> FileMetadata:
    response = client.post(
        "/sync/get_metadata",
        json={
            "path_like": path.as_posix(),
        },
    )

    response_data = handle_json_response("/sync/get_metadata", response)

    if len(response_data) == 0:
        raise SyftNotFound(f"[/sync/get_metadata] not found on server: {path}")
    return FileMetadata(**response_data[0])


def get_diff(client: httpx.Client, path: Path, signature: bytes) -> DiffResponse:
    response = client.post(
        "/sync/get_diff",
        json={
            "path": str(path),
            "signature": base64.b85encode(signature).decode("utf-8"),
        },
    )

    response_data = handle_json_response("/sync/get_diff", response)
    return DiffResponse(**response_data)


def apply_diff(client: httpx.Client, path: Path, diff: bytes, expected_hash: str) -> ApplyDiffResponse:
    response = client.post(
        "/sync/apply_diff",
        json={
            "path": str(path),
            "diff": base64.b85encode(diff).decode("utf-8"),
            "expected_hash": expected_hash,
        },
    )

    response_data = handle_json_response("/sync/apply_diff", response)
    return ApplyDiffResponse(**response_data)


def delete(client: httpx.Client, path: Path) -> None:
    response = client.post(
        "/sync/delete",
        json={
            "path": str(path),
        },
    )

    response.raise_for_status()


def create(client: httpx.Client, path: Path, data: bytes) -> None:
    response = client.post("/sync/create", files={"file": (str(path), data, "text/plain")})
    response = handle_json_response("/sync/create", response)
    return


def download(client: httpx.Client, path: Path) -> bytes:
    response = client.post(
        "/sync/download",
        json={
            "path": str(path),
        },
    )

    if response.status_code != 200:
        raise SyftNotFound(f"[/sync/download] not found on server: {path}, {response.text}")

    return response.content



import json
from pathlib import Path

import faker
from fastapi.testclient import TestClient

from syftbox.client.plugins.sync.manager import SyncManager, SyncQueueItem
from syftbox.client.plugins.sync.sync import is_permission_file
from syftbox.client.utils.dir_tree import DirTree, create_dir_tree
from syftbox.lib import Client
from syftbox.lib.lib import ClientConfig, SyftPermission
from syftbox.server.settings import ServerSettings

fake = faker.Faker()


def create_random_file(client_config: ClientConfig, sub_path: str = "") -> Path:
    relative_path = Path(sub_path) / fake.file_name(extension="json")
    file_path = client_config.datasite_path / relative_path
    content = {"body": fake.text()}
    file_path.write_text(json.dumps(content))

    path_in_datasite = file_path.relative_to(client_config.sync_folder)
    return path_in_datasite


def assert_files_not_on_datasite(datasite: ClientConfig, files: list[Path]):
    for file in files:
        assert not (datasite.sync_folder / file).exists(), f"File {file} exists on datasite {datasite.email}"


def assert_files_on_datasite(datasite: ClientConfig, files: list[Path]):
    for file in files:
        assert (datasite.sync_folder / file).exists(), f"File {file} does not exist on datasite {datasite.email}"


def assert_files_on_server(server_client: TestClient, files: list[Path]):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    for file in files:
        assert (server_settings.snapshot_folder / file).exists(), f"File {file} does not exist on server"


def assert_dirtree_exists(base_path: Path, tree: DirTree) -> None:
    for name, content in tree.items():
        local_path = base_path / name

        if isinstance(content, str):
            assert local_path.read_text() == content
        elif isinstance(content, SyftPermission):
            assert json.loads(local_path.read_text()) == content.to_dict()
        elif isinstance(content, dict):
            assert local_path.is_dir()
            assert_dirtree_exists(local_path, content)


def test_get_datasites(datasite_1: Client, datasite_2: Client):
    emails = {datasite_1.email, datasite_2.email}
    sync_service = SyncManager(datasite_1)
    sync_service2 = SyncManager(datasite_2)
    sync_service.run_single_thread()
    sync_service2.run_single_thread()

    datasites = sync_service.get_datasites()
    assert {datasites[0].email, datasites[1].email} == emails


def test_enqueue_changes(datasite_1: Client):
    sync_service = SyncManager(datasite_1)
    datasites = sync_service.get_datasites()

    out_of_sync_permissions, out_of_sync_files = datasites[0].get_out_of_sync_files()
    num_files_after_setup = len(out_of_sync_files) + len(out_of_sync_permissions)

    # Create two files in datasite_1
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
            "large.txt": fake.text(max_nb_chars=1000),
            "small.txt": fake.text(max_nb_chars=10),
        },
    }
    create_dir_tree(Path(datasite_1.datasite_path), tree)
    out_of_sync_permissions, out_of_sync_files = datasites[0].get_out_of_sync_files()
    num_out_of_sync_files = len(out_of_sync_files) + len(out_of_sync_permissions)
    # 3 new files
    assert num_out_of_sync_files - num_files_after_setup == 3

    # Enqueue the changes + verify order
    for change in out_of_sync_permissions + out_of_sync_files:
        sync_service.enqueue(change)

    items_from_queue: list[SyncQueueItem] = []
    while not sync_service.queue.empty():
        items_from_queue.append(sync_service.queue.get())

    should_be_permissions = items_from_queue[: len(out_of_sync_permissions)]
    should_be_files = items_from_queue[len(out_of_sync_permissions) :]

    assert all(is_permission_file(item.data.path) for item in should_be_permissions)
    assert all(not is_permission_file(item.data.path) for item in should_be_files)

    for item in should_be_files:
        print(item.priority, item.data)


def test_create_file(server_client: TestClient, datasite_1: Client, datasite_2: Client):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    sync_service = SyncManager(datasite_1)

    # Create a file in datasite_1
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
            "file.txt": fake.text(max_nb_chars=1000),
        },
    }
    create_dir_tree(Path(datasite_1.datasite_path), tree)

    # changes are pushed to server
    sync_service.run_single_thread()

    # check if no changes are left
    for datasite in sync_service.get_datasites():
        out_of_sync_permissions, out_of_sync_files = datasite.get_out_of_sync_files()
        assert not out_of_sync_files
        assert not out_of_sync_permissions

    # check if file exists on server
    print(datasite_2.sync_folder)
    datasite_snapshot = server_settings.snapshot_folder / datasite_1.email
    assert_dirtree_exists(datasite_snapshot, tree)

    # check if file exists on datasite_2
    sync_client_2 = SyncManager(datasite_2)
    sync_client_2.run_single_thread()
    datasite_states = sync_client_2.get_datasites()
    ds1_state = datasite_states[0]
    assert ds1_state.email == datasite_1.email

    print(ds1_state.get_out_of_sync_files())

    print(f"datasites {[d.email for d in sync_client_2.get_datasites()]}")
    sync_client_2.run_single_thread()

    assert_files_on_datasite(datasite_2, [Path(datasite_1.email) / "folder1" / "file.txt"])


def test_modify(server_client: TestClient, datasite_1: Client):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    sync_service_1 = SyncManager(datasite_1)

    # Setup initial state
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_write(datasite_1.email),
            "file.txt": "content",
        },
    }
    create_dir_tree(Path(datasite_1.datasite_path), tree)
    sync_service_1.run_single_thread()

    # modify the file
    file_path = datasite_1.datasite_path / "folder1" / "file.txt"
    new_content = "modified"
    file_path.write_text(new_content)
    assert file_path.read_text() == new_content

    sync_service_1.run_single_thread()

    assert file_path.read_text() == new_content
    assert (server_settings.snapshot_folder / datasite_1.email / "folder1" / "file.txt").read_text() == new_content


def test_modify_with_conflict(server_client: TestClient, datasite_1: Client, datasite_2: Client):
    sync_service_1 = SyncManager(datasite_1)
    sync_service_2 = SyncManager(datasite_2)

    # Setup initial state
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_write(datasite_1.email),
            "file.txt": "content1",
        },
    }
    create_dir_tree(Path(datasite_1.datasite_path), tree)
    sync_service_1.run_single_thread()
    sync_service_2.run_single_thread()

    # modify the file both clients
    file_path_1 = datasite_1.datasite_path / "folder1" / "file.txt"
    new_content_1 = "modified1"
    file_path_1.write_text(new_content_1)

    file_path_2 = Path(datasite_2.sync_folder) / datasite_1.email / "folder1" / "file.txt"
    new_content_2 = "modified2"
    file_path_2.write_text(new_content_2)

    assert new_content_1 != new_content_2
    assert file_path_1.read_text() == new_content_1
    assert file_path_2.read_text() == new_content_2

    # first to server wins
    sync_service_1.run_single_thread()
    sync_service_2.run_single_thread()

    assert file_path_1.read_text() == new_content_1
    assert file_path_2.read_text() == new_content_1

    # modify again, 2 syncs first
    new_content_1 = fake.text(max_nb_chars=1000)
    new_content_2 = fake.text(max_nb_chars=1000)
    file_path_1.write_text(new_content_1)
    file_path_2.write_text(new_content_2)
    assert new_content_1 != new_content_2

    assert file_path_1.read_text() == new_content_1
    assert file_path_2.read_text() == new_content_2

    sync_service_2.run_single_thread()
    sync_service_1.run_single_thread()

    assert file_path_1.read_text() == new_content_2
    assert file_path_2.read_text() == new_content_2


def test_delete_file(server_client: TestClient, datasite_1: Client, datasite_2: Client):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    sync_service_1 = SyncManager(datasite_1)
    sync_service_2 = SyncManager(datasite_2)

    # Setup initial state
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_write(datasite_1.email),
            "file.txt": fake.text(max_nb_chars=1000),
        },
    }
    create_dir_tree(Path(datasite_1.datasite_path), tree)
    sync_service_1.run_single_thread()
    sync_service_2.run_single_thread()

    # delete the file
    file_path = datasite_1.datasite_path / "folder1" / "file.txt"
    file_path.unlink()

    sync_service_1.run_single_thread()

    # file is deleted on server
    assert (server_settings.snapshot_folder / datasite_1.email / "folder1" / "file.txt").exists() is False

    sync_service_2.run_single_thread()
    assert (datasite_2.datasite_path / datasite_1.email / "folder1" / "file.txt").exists() is False

    # Check if the metadata is gone
    remote_state_1 = sync_service_1.get_datasites()[0].get_remote_state()
    remote_paths = {metadata.path for metadata in remote_state_1}
    assert Path(datasite_1.email) / "folder1" / "file.txt" not in remote_paths



import base64
import enum
from datetime import datetime
from pathlib import Path
from typing import Any, Optional

from pydantic import BaseModel, Field


class DiffRequest(BaseModel):
    path: Path
    signature: str

    @property
    def signature_bytes(self) -> bytes:
        return base64.b85decode(self.signature)


class DiffResponse(BaseModel):
    path: Path
    diff: str
    hash: str

    @property
    def diff_bytes(self) -> bytes:
        return base64.b85decode(self.diff)


class SignatureError(str, enum.Enum):
    FILE_NOT_FOUND = "FILE_NOT_FOUND"
    FILE_NOT_WRITEABLE = "FILE_NOT_WRITEABLE"
    FILE_NOT_READABLE = "FILE_NOT_READABLE"
    NOT_A_FILE = "NOT_A_FILE"


class SignatureResponse(BaseModel):
    path: str
    signature: Optional[str] = None
    error: Optional[SignatureError] = None


class FileMetadataRequest(BaseModel):
    path_like: str = Field(description="Path to search for files, uses SQL LIKE syntax")


class FileRequest(BaseModel):
    path: str = Field(description="Path to search for files, uses SQL LIKE syntax")


class ApplyDiffRequest(BaseModel):
    path: str
    diff: str
    expected_hash: str

    @property
    def diff_bytes(self) -> bytes:
        return base64.b85decode(self.diff)


class ApplyDiffResponse(BaseModel):
    path: str
    current_hash: str
    previous_hash: str


class FileMetadata(BaseModel):
    path: Path
    hash: str
    signature: str
    file_size: int = 0  # limit file sizes?
    last_modified: datetime

    @property
    def signature_bytes(self) -> bytes:
        return base64.b85decode(self.signature)

    @property
    def hash_bytes(self) -> bytes:
        return base64.b85decode(self.hash)

    def __eq__(self, value: Any):
        if not isinstance(value, FileMetadata):
            return False
        return self.path == value.path and self.hash == value.hash


class SyncLog(BaseModel):
    path: Path
    method: str  # pull or push
    status: str  # success or failure
    timestamp: datetime
    requesting_user: str



from fastapi.testclient import TestClient


def test_register(client: TestClient):
    data = {"email": "test@example.com"}
    response = client.post("/register", json=data)
    assert response.status_code == 200
    assert "token" in response.json()
